@@ -14,13 +14,13 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmlog.h>
 
-#include "lib/misc.h"
-#include "lib/rpmplugins.h"
-#include "lib/rpmte_internal.h"
+#include "misc.h"
+#include "rpmplugins.h"
+#include "rpmte_internal.h"
 /* strpool-related interfaces */
-#include "lib/rpmfi_internal.h"
-#include "lib/rpmds_internal.h"
-#include "lib/rpmts_internal.h"
+#include "rpmfi_internal.h"
+#include "rpmds_internal.h"
+#include "rpmts_internal.h"
 
 #include "debug.h"
 
@@ -58,7 +58,6 @@ struct rpmte_s {
     rpmds supplements;		/*!< Supplements: dependencies. */
     rpmds enhances;		/*!< Enhances: dependencies. */
     rpmfiles files;		/*!< File information. */
-    rpmfi fi;			/*!< File iterator (backwards compat) */
     rpmps probs;		/*!< Problems (relocations) */
     rpmts ts;			/*!< Parent transaction */
 
@@ -76,6 +75,8 @@ struct rpmte_s {
 
 #define RPMTE_HAVE_PRETRANS	(1 << 0)
 #define RPMTE_HAVE_POSTTRANS	(1 << 1)
+#define RPMTE_HAVE_PREUNTRANS	(1 << 2)
+#define RPMTE_HAVE_POSTUNTRANS	(1 << 3)
     int transscripts;		/*!< pre/posttrans script existence */
     int failed;			/*!< (parent) install/erase failed */
 
@@ -208,6 +209,12 @@ static int addTE(rpmte p, Header h, fnpyKey key, rpmRelocation * relocs)
     p->transscripts |= (headerIsEntry(h, RPMTAG_POSTTRANS) ||
 			 headerIsEntry(h, RPMTAG_POSTTRANSPROG)) ?
 			RPMTE_HAVE_POSTTRANS : 0;
+    p->transscripts |= (headerIsEntry(h, RPMTAG_PREUNTRANS) ||
+			 headerIsEntry(h, RPMTAG_PREUNTRANSPROG)) ?
+			RPMTE_HAVE_PREUNTRANS : 0;
+    p->transscripts |= (headerIsEntry(h, RPMTAG_POSTUNTRANS) ||
+			 headerIsEntry(h, RPMTAG_POSTUNTRANSPROG)) ?
+			RPMTE_HAVE_POSTUNTRANS : 0;
 
     rpmteColorDS(p, RPMTAG_PROVIDENAME);
     rpmteColorDS(p, RPMTAG_REQUIRENAME);
@@ -243,7 +250,6 @@ rpmte rpmteFree(rpmte te)
 	free(te->NEVRA);
 
 	fdFree(te->fd);
-	rpmfiFree(te->fi);
 	rpmfilesFree(te->files);
 	headerFree(te->h);
 	rpmfsFree(te->fs);
@@ -259,7 +265,7 @@ rpmte rpmteFree(rpmte te)
 rpmte rpmteNew(rpmts ts, Header h, rpmElementType type, fnpyKey key,
 	       rpmRelocation * relocs, int addop)
 {
-    rpmte p = xcalloc(1, sizeof(*p));
+    rpmte p = (rpmte)xcalloc(1, sizeof(*p));
     p->ts = ts;
     p->type = type;
     p->addop = addop;
@@ -301,8 +307,7 @@ Header rpmteSetHeader(rpmte te, Header h)
 
 rpmElementType rpmteType(rpmte te)
 {
-    /* XXX returning negative for unsigned type */
-    return (te != NULL ? te->type : -1);
+    return te->type;
 }
 
 const char * rpmteN(rpmte te)
@@ -471,22 +476,10 @@ rpmds rpmteDS(rpmte te, rpmTagVal tag)
 void rpmteCleanFiles(rpmte te)
 {
     if (te != NULL)  {
-	te->fi = rpmfiFree(te->fi);
 	te->files = rpmfilesFree(te->files);
     }
 }
 
-rpmfi rpmteFI(rpmte te)
-{
-    if (te == NULL)
-	return NULL;
-
-    if (te->fi == NULL)
-	te->fi = rpmfilesIter(te->files, RPMFI_ITER_FWD);
-
-    return te->fi; /* XXX take fi reference here? */
-}
-
 rpmfiles rpmteFiles(rpmte te)
 {
     return (te != NULL) ? rpmfilesLink(te->files) : NULL;
@@ -496,7 +489,7 @@ static void rpmteColorDS(rpmte te, rpmTag tag)
 {
     rpmfi fi;
     rpmds ds = rpmteDS(te, tag);
-    char deptype = 'R';
+    char deptype = rpmdsD(ds);
     char mydt;
     const uint32_t * ddict;
     rpm_color_t * colors;
@@ -505,22 +498,10 @@ static void rpmteColorDS(rpmte te, rpmTag tag)
     unsigned ix;
     int ndx, i;
 
-    if (!(te && (Count = rpmdsCount(ds)) > 0 && rpmfilesFC(te->files) > 0))
+    if (!(te && deptype && (Count = rpmdsCount(ds)) > 0 && rpmfilesFC(te->files) > 0))
 	return;
 
-    switch (tag) {
-    default:
-	return;
-	break;
-    case RPMTAG_PROVIDENAME:
-	deptype = 'P';
-	break;
-    case RPMTAG_REQUIRENAME:
-	deptype = 'R';
-	break;
-    }
-
-    colors = xcalloc(Count, sizeof(*colors));
+    colors = (rpm_color_t *)xcalloc(Count, sizeof(*colors));
 
     /* Calculate dependency color. */
     fi = rpmfilesIter(te->files, RPMFI_ITER_FWD);
@@ -568,7 +549,7 @@ static Header rpmteDBHeader(rpmte te)
 static Header rpmteFDHeader(rpmte te)
 {
     Header h = NULL;
-    te->fd = rpmtsNotify(te->ts, te, RPMCALLBACK_INST_OPEN_FILE, 0, 0);
+    te->fd = (FD_t)rpmtsNotify(te->ts, te, RPMCALLBACK_INST_OPEN_FILE, 0, 0);
     if (te->fd != NULL) {
 	rpmVSFlags ovsflags;
 	rpmRC pkgrc;
@@ -597,18 +578,21 @@ static int rpmteOpen(rpmte te, int reload_fi)
     if (te == NULL || te->ts == NULL || rpmteFailed(te))
 	goto exit;
 
-    rpmteSetHeader(te, NULL);
-
     switch (rpmteType(te)) {
     case TR_ADDED:
 	h = rpmteDBInstance(te) ? rpmteDBHeader(te) : rpmteFDHeader(te);
 	break;
     case TR_REMOVED:
+	if (te->h) {
+	    h = headerLink(te->h);
+	    break;
+	}
     case TR_RPMDB:
     case TR_RESTORED:
 	h = rpmteDBHeader(te);
     	break;
     }
+
     if (h != NULL) {
 	if (reload_fi) {
 	    /* This can fail if we get a different, bad header from callback */
@@ -628,6 +612,7 @@ exit:
 
 static int rpmteClose(rpmte te, int reset_fi)
 {
+    int keephdr = 0;
     if (te == NULL || te->ts == NULL)
 	return 0;
 
@@ -639,12 +624,15 @@ static int rpmteClose(rpmte te, int reset_fi)
 	}
 	break;
     case TR_REMOVED:
+	if (te->transscripts & RPMTE_HAVE_POSTUNTRANS)
+	    keephdr = 1;
     case TR_RPMDB:
     case TR_RESTORED:
 	/* eventually we'll want notifications for erase open too */
 	break;
     }
-    rpmteSetHeader(te, NULL);
+    if (!keephdr)
+	rpmteSetHeader(te, NULL);
     if (reset_fi) {
 	rpmteCleanFiles(te);
     }
@@ -690,11 +678,16 @@ int rpmteFailed(rpmte te)
 
 int rpmteHaveTransScript(rpmte te, rpmTagVal tag)
 {
-    int rc = 0;
+    /* We only filter pre/post transaction scripts */
+    int rc = 1;
     if (tag == RPMTAG_PRETRANS) {
 	rc = (te->transscripts & RPMTE_HAVE_PRETRANS);
     } else if (tag == RPMTAG_POSTTRANS) {
 	rc = (te->transscripts & RPMTE_HAVE_POSTTRANS);
+    } else if (tag == RPMTAG_PREUNTRANS) {
+	rc = (te->transscripts & RPMTE_HAVE_PREUNTRANS);
+    } else if (tag == RPMTAG_POSTUNTRANS) {
+	rc = (te->transscripts & RPMTE_HAVE_POSTUNTRANS);
     }
     return rc;
 }
@@ -814,11 +807,8 @@ int rpmteProcess(rpmte te, pkgGoal goal, int num)
     int failed = 1;
 
     /* Dont bother opening for elements without pre/posttrans scripts */
-    if (goal == PKG_PRETRANS || goal == PKG_POSTTRANS) {
-	if (!rpmteHaveTransScript(te, goal)) {
-	    return 0;
-	}
-    }
+    if (!rpmteHaveTransScript(te, goal))
+	return 0;
 
     if (rpmteOpen(te, reset_fi)) {
 	if (!scriptstage) {
