@@ -6,22 +6,22 @@
 #include <rpm/rpmfi.h>
 #include <stdlib.h>
 
-#include "lib/rpmtriggers.h"
-#include "lib/rpmts_internal.h"
-#include "lib/rpmdb_internal.h"
-#include "lib/rpmds_internal.h"
-#include "lib/rpmfi_internal.h"
-#include "lib/rpmte_internal.h"
-#include "lib/rpmchroot.h"
+#include "rpmtriggers.h"
+#include "rpmts_internal.h"
+#include "rpmdb_internal.h"
+#include "rpmds_internal.h"
+#include "rpmfi_internal.h"
+#include "rpmte_internal.h"
+#include "rpmchroot.h"
 
 #define TRIGGER_PRIORITY_BOUND 10000
 
 rpmtriggers rpmtriggersCreate(unsigned int hint)
 {
-    rpmtriggers triggers = xmalloc(sizeof(struct rpmtriggers_s));
+    rpmtriggers triggers = (rpmtriggers)xmalloc(sizeof(struct rpmtriggers_s));
     triggers->count = 0;
     triggers->alloced = hint;
-    triggers->triggerInfo = xmalloc(sizeof(struct triggerInfo_s) *
+    triggers->triggerInfo = (struct triggerInfo_s *)xmalloc(sizeof(struct triggerInfo_s) *
 				    triggers->alloced);
     return triggers;
 }
@@ -51,7 +51,8 @@ static void rpmtriggersAdd(rpmtriggers trigs, unsigned int hdrNum,
 
 static int trigCmp(const void *a, const void *b)
 {
-    const struct triggerInfo_s *trigA = a, *trigB = b;
+    const struct triggerInfo_s *trigA = (const struct triggerInfo_s *)a;
+    const struct triggerInfo_s *trigB = (const struct triggerInfo_s *)b;
 
     if (trigA->priority < trigB->priority)
 	return 1;
@@ -97,23 +98,32 @@ static void rpmtriggersSortAndUniq(rpmtriggers trigs)
     }
 }
 
+static unsigned int getTrigPriority(Header h, rpmTagVal prioTag, int tix)
+{
+    unsigned int priority = RPMTRIGGER_DEFAULT_PRIORITY;
+    struct rpmtd_s priorities;
+
+    if (headerGet(h, prioTag, &priorities, HEADERGET_MINMEM)) {
+	if (rpmtdSetIndex(&priorities, tix) >= 0)
+	    priority = rpmtdGetNumber(&priorities);
+    }
+    return priority;
+}
+
 static void addTriggers(rpmts ts, Header trigH, rpmsenseFlags filter,
 			const char *prefix)
 {
     int tix = 0;
     rpmds ds;
     rpmds triggers = rpmdsNew(trigH, RPMTAG_TRANSFILETRIGGERNAME, 0);
+    rpmTagVal prioTag = RPMTAG_TRANSFILETRIGGERPRIORITIES;
 
     while ((ds = rpmdsFilterTi(triggers, tix))) {
-	if ((rpmdsNext(ds) >= 0) && (rpmdsFlags(ds) & filter) &&
-		strcmp(prefix, rpmdsN(ds)) == 0) {
-	    struct rpmtd_s priorities;
-
-	    if (headerGet(trigH, RPMTAG_TRANSFILETRIGGERPRIORITIES,
-			&priorities, HEADERGET_MINMEM)) {
-		rpmtdSetIndex(&priorities, tix);
+	while (rpmdsNext(ds) >= 0) {
+	    if ((rpmdsFlags(ds) & filter) && strcmp(prefix, rpmdsN(ds)) == 0) {
+		unsigned int priority = getTrigPriority(trigH, prioTag, tix);
 		rpmtriggersAdd(ts->trigs2run, headerGetInstance(trigH),
-				tix, *rpmtdGetUint32(&priorities));
+				    tix, priority);
 	    }
 	}
 	rpmdsFree(ds);
@@ -162,6 +172,8 @@ int runPostUnTransFileTrigs(rpmts ts)
 {
     int i;
     Header trigH;
+    const char * trigName = NULL;
+    int arg1 = 0;
     struct rpmtd_s installPrefixes;
     rpmScript script;
     rpmtriggers trigs = ts->trigs2run;
@@ -186,7 +198,11 @@ int runPostUnTransFileTrigs(rpmts ts)
 	headerGet(trigH, RPMTAG_INSTPREFIXES, &installPrefixes,
 		HEADERGET_ALLOC|HEADERGET_ARGV);
 
-	nerrors += runScript(ts, NULL, trigH, installPrefixes.data, script, 0, -1);
+	trigName = headerGetString(trigH, RPMTAG_NAME);
+	arg1 = rpmdbCountPackages(rpmtsGetRdb(ts), trigName);
+
+	ARGV_const_t prefixes = (ARGV_const_t)installPrefixes.data;
+	nerrors += runScript(ts, NULL, trigH, prefixes, script, arg1, -1);
 	rpmtdFreeData(&installPrefixes);
 	rpmScriptFree(script);
 	headerFree(trigH);
@@ -195,50 +211,6 @@ int runPostUnTransFileTrigs(rpmts ts)
     return nerrors;
 }
 
-/*
- * Get files from next package from match iterator. If files are
- * available in memory then don't read them from rpmdb.
- */
-static rpmfiles rpmtsNextFiles(rpmts ts, rpmdbMatchIterator mi)
-{
-    Header h;
-    rpmte *te;
-    rpmfiles files = NULL;
-    rpmstrPool pool = ts->members->pool;
-    int ix;
-    unsigned int offset;
-
-    ix = rpmdbGetIteratorIndex(mi);
-    if (ix < rpmdbGetIteratorCount(mi)) {
-	offset = rpmdbGetIteratorOffsetFor(mi, ix);
-	if (packageHashGetEntry(ts->members->removedPackages, offset,
-				&te, NULL, NULL)) {
-	    /* Files are available in memory */
-	    files  = rpmteFiles(te[0]);
-	}
-
-	if (packageHashGetEntry(ts->members->installedPackages, offset,
-				&te, NULL, NULL)) {
-	    /* Files are available in memory */
-	    files  = rpmteFiles(te[0]);
-	}
-    }
-
-    if (files) {
-	rpmdbSetIteratorIndex(mi, ix + 1);
-    } else {
-	/* Files are not available in memory. Read them from rpmdb */
-	h = rpmdbNextIterator(mi);
-	if (h) {
-	    files = rpmfilesNew(pool, h, RPMTAG_BASENAMES,
-				RPMFI_FLAGS_FILETRIGGER);
-	}
-    }
-
-    return files;
-}
-
-
 typedef struct matchFilesIter_s {
     rpmts ts;
     rpmds rpmdsTrigger;
@@ -246,13 +218,59 @@ typedef struct matchFilesIter_s {
     rpmfi fi;
     rpmfs fs;
     const char *pfx;
+    const char *pkgname;
     rpmdbMatchIterator pi;
     packageHash tranPkgs;
 } *matchFilesIter;
 
+/*
+ * Get files from next package from match iterator. If files are
+ * available in memory then don't read them from rpmdb.
+ */
+static rpmfiles rpmtsNextFiles(matchFilesIter mfi)
+{
+    Header h;
+    rpmte *te;
+    rpmfiles files = NULL;
+    rpmstrPool pool = mfi->ts->members->pool;
+    int ix;
+    unsigned int offset;
+
+    ix = rpmdbGetIteratorIndex(mfi->pi);
+    if (ix < rpmdbGetIteratorCount(mfi->pi)) {
+	offset = rpmdbGetIteratorOffsetFor(mfi->pi, ix);
+	if (packageHashGetEntry(mfi->ts->members->removedPackages, offset,
+				&te, NULL, NULL)) {
+	    /* Files are available in memory */
+	    files  = rpmteFiles(te[0]);
+	}
+
+	if (packageHashGetEntry(mfi->ts->members->installedPackages, offset,
+				&te, NULL, NULL)) {
+	    /* Files are available in memory */
+	    files  = rpmteFiles(te[0]);
+	}
+    }
+
+    if (files) {
+	rpmdbSetIteratorIndex(mfi->pi, ix + 1);
+	mfi->pkgname = rpmteN(te[0]);
+    } else {
+	/* Files are not available in memory. Read them from rpmdb */
+	h = rpmdbNextIterator(mfi->pi);
+	if (h) {
+	    files = rpmfilesNew(pool, h, RPMTAG_BASENAMES,
+				RPMFI_FLAGS_FILETRIGGER);
+	    mfi->pkgname = headerGetString(h, RPMTAG_NAME);
+	}
+    }
+
+    return files;
+}
+
 static matchFilesIter matchFilesIterator(rpmds trigger, rpmfiles files, rpmte te)
 {
-    matchFilesIter mfi = xcalloc(1, sizeof(*mfi));
+    matchFilesIter mfi = (matchFilesIter)xcalloc(1, sizeof(*mfi));
     rpmdsInit(trigger);
     mfi->rpmdsTrigger = trigger;
     mfi->files = rpmfilesLink(files);
@@ -263,7 +281,7 @@ static matchFilesIter matchFilesIterator(rpmds trigger, rpmfiles files, rpmte te
 static matchFilesIter matchDBFilesIterator(rpmds trigger, rpmts ts,
 					    int inTransaction)
 {
-    matchFilesIter mfi = xcalloc(1, sizeof(*mfi));
+    matchFilesIter mfi = (matchFilesIter)xcalloc(1, sizeof(*mfi));
     rpmsenseFlags sense;
 
     rpmdsSetIx(trigger, 0);
@@ -321,7 +339,7 @@ static const char *matchFilesNext(matchFilesIter mfi)
 	    /* If we are done with current mfi->fi, create mfi->fi for next package */
 	    rpmfilesFree(mfi->files);
 	    rpmfiFree(mfi->fi);
-	    mfi->files = rpmtsNextFiles(mfi->ts, mfi->pi);
+	    mfi->files = rpmtsNextFiles(mfi);
 	    mfi->fi = rpmfilesFindPrefix(mfi->files, mfi->pfx);
 	    if (mfi->files)
 		continue;
@@ -379,7 +397,7 @@ static matchFilesIter matchFilesIteratorFree(matchFilesIter mfi)
  */
 static int runHandleTriggersInPkg(rpmts ts, rpmte te, Header h,
 				rpmsenseFlags sense, rpmscriptTriggerModes tm,
-				int searchMode, int ti)
+				int searchMode, int ti, int arg1, int arg2)
 {
     int nerrors = 0;
     rpmds rpmdsTriggers, rpmdsTrigger;
@@ -387,7 +405,7 @@ static int runHandleTriggersInPkg(rpmts ts, rpmte te, Header h,
     matchFilesIter mfi = NULL;
     rpmScript script;
     struct rpmtd_s installPrefixes;
-    char *(*inputFunc)(void *);
+    nextfilefunc inputFunc;
 
     rpmdsTriggers = rpmdsNew(h, triggerDsTag(tm), 0);
     rpmdsTrigger = rpmdsFilterTi(rpmdsTriggers, ti);
@@ -430,11 +448,15 @@ static int runHandleTriggersInPkg(rpmts ts, rpmte te, Header h,
 	     * that will be passed as stdin to trigger script. To get
 	     * these data from lua script function rpm.input() can be used.
 	     */
-	    inputFunc = (char *(*)(void *)) matchFilesNext;
+	    inputFunc = (nextfilefunc) matchFilesNext;
 	    rpmScriptSetNextFileFunc(script, inputFunc, mfi);
 
-	    nerrors += runScript(ts, NULL, h, installPrefixes.data,
-				script, 0, -1);
+	    /* Compute our own arg2 if asked */
+	    if (arg2 < 0 && tm == RPMSCRIPT_FILETRIGGER && mfi->pkgname)
+		arg2 = rpmdbCountPackages(rpmtsGetRdb(ts), mfi->pkgname);
+
+	    ARGV_const_t prefixes = (ARGV_const_t)installPrefixes.data;
+	    nerrors += runScript(ts, NULL, h, prefixes, script, arg1, arg2);
 	    rpmtdFreeData(&installPrefixes);
 	    rpmScriptFree(script);
 	}
@@ -484,7 +506,7 @@ static int matchFilesInTran(rpmts ts, rpmte te, const char *pfx,
     return rc;
 }
 
-rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
+rpmRC runFileTriggers(rpmts ts, rpmte te, int arg2, rpmsenseFlags sense,
 			rpmscriptTriggerModes tm, int priorityClass)
 {
     int nerrors = 0, i;
@@ -493,6 +515,8 @@ rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
     char *pfx;
     size_t keylen;
     Header trigH;
+    const char * trigName = NULL;
+    int arg1 = 0;
     int (*matchFunc)(rpmts, rpmte, const char*, rpmsenseFlags sense);
     rpmTagVal priorityTag;
     rpmtriggers triggers = rpmtriggersCreate(10);
@@ -506,20 +530,18 @@ rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
 	priorityTag = RPMTAG_TRANSFILETRIGGERPRIORITIES;
     }
 
-    ii = rpmdbIndexIteratorInit(rpmtsGetRdb(ts), triggerDsTag(tm));
+    ii = rpmdbIndexIteratorInit(rpmtsGetRdb(ts), (rpmDbiTag)triggerDsTag(tm));
 
     /* Loop over all file triggers in rpmdb */
     while ((rpmdbIndexIteratorNext(ii, &key, &keylen)) == 0) {
-	pfx = xmalloc(keylen + 1);
+	pfx = (char *)xmalloc(keylen + 1);
 	memcpy(pfx, key, keylen);
 	pfx[keylen] = '\0';
 
 	/* Check if file trigger is fired by any file in ts/te */
 	if (matchFunc(ts, te, pfx, sense)) {
 	    for (i = 0; i < rpmdbIndexIteratorNumPkgs(ii); i++) {
-		struct rpmtd_s priorities;
 		unsigned int priority = 0;
-		unsigned int *priority_ptr;
 		unsigned int offset = rpmdbIndexIteratorPkgOffset(ii, i);
 		unsigned int tix = rpmdbIndexIteratorTagNum(ii, i);
 
@@ -535,11 +557,7 @@ rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
 
 		/* Get priority of trigger from header */
 		trigH = rpmdbGetHeaderAt(rpmtsGetRdb(ts), offset);
-		headerGet(trigH, priorityTag, &priorities, HEADERGET_MINMEM);
-		rpmtdSetIndex(&priorities, tix);
-		priority_ptr = rpmtdGetUint32(&priorities);
-		if (priority_ptr)
-		    priority = *priority_ptr;
+		priority = getTrigPriority(trigH, priorityTag, tix);
 		headerFree(trigH);
 
 		/* Store file trigger in array */
@@ -564,12 +582,17 @@ rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
 	}
 
 	trigH = rpmdbGetHeaderAt(rpmtsGetRdb(ts), triggers->triggerInfo[i].hdrNum);
+	trigName = headerGetString(trigH, RPMTAG_NAME);
+	arg1 = rpmdbCountPackages(rpmtsGetRdb(ts), trigName);
+
 	if (tm == RPMSCRIPT_FILETRIGGER)
 	    nerrors += runHandleTriggersInPkg(ts, te, trigH, sense, tm, 0,
-						triggers->triggerInfo[i].tix);
+						triggers->triggerInfo[i].tix,
+						arg1, arg2);
 	else
 	    nerrors += runHandleTriggersInPkg(ts, te, trigH, sense, tm, 1,
-						triggers->triggerInfo[i].tix);
+						triggers->triggerInfo[i].tix,
+						arg1, arg2);
 	headerFree(trigH);
     }
     rpmtriggersFree(triggers);
@@ -577,7 +600,7 @@ rpmRC runFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
     return (nerrors == 0) ? RPMRC_OK : RPMRC_FAIL;
 }
 
-rpmRC runImmedFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
+rpmRC runImmedFileTriggers(rpmts ts, rpmte te, int arg1, rpmsenseFlags sense,
 			    rpmscriptTriggerModes tm, int priorityClass)
 {
     int nerrors = 0;
@@ -598,9 +621,11 @@ rpmRC runImmedFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
     triggers = rpmtriggersCreate(triggersCount);
 
     for (i = 0; i < triggersCount; i++) {
-	rpmtdSetIndex(&priorities, i);
+	unsigned int priority = RPMTRIGGER_DEFAULT_PRIORITY;
+	if (rpmtdSetIndex(&priorities, i) >= 0)
+	    priority = rpmtdGetNumber(&priorities);
 	/* Offset is not important, all triggers are from the same package */
-	rpmtriggersAdd(triggers, 0, i, *rpmtdGetUint32(&priorities));
+	rpmtriggersAdd(triggers, 0, i, priority);
     }
 
     /* Sort triggers by priority, offset, trigger index */
@@ -616,7 +641,8 @@ rpmRC runImmedFileTriggers(rpmts ts, rpmte te, rpmsenseFlags sense,
 	}
 
 	nerrors += runHandleTriggersInPkg(ts, te, trigH, sense, tm, 2,
-					    triggers->triggerInfo[i].tix);
+					    triggers->triggerInfo[i].tix,
+					    arg1, -1);
     }
     rpmtriggersFree(triggers);
     headerFree(trigH);
